
#include <SPI.h>


#define CS_PIN PA4

enum {
  SFLASH_CMD_PAGE_PROGRAM = 0x02,
  SFLASH_CMD_READ = 0x03,      // Single Read
  SFLASH_CMD_FAST_READ = 0x0B, // Fast Read
  SFLASH_CMD_WRITE_ENABLE = 0x06,
  SFLASH_CMD_WRITE_DISABLE = 0x04,
  SFLASH_CMD_ERASE_SECTOR = 0x20,
  SFLASH_CMD_ERASE_BLOCK = 0xD8,
  SFLASH_CMD_ERASE_CHIP = 0xC7,
};

enum {
  SFLASH_BLOCK_SIZE = 64 * 1024UL,
  SFLASH_SECTOR_SIZE = 4 * 1024,
  SFLASH_PAGE_SIZE = 256,
};


  uint8_t _cmd_read = SFLASH_CMD_READ;
  uint8_t _addr_len = 3; // work with most device if not set
  SPIClass * _spi = &SPI;

// Execute a single byte command e.g Reset, Write Enable
void Flash_runCommand(uint8_t command)
{
  digitalWrite(CS_PIN, LOW);
  _spi->transfer(command);
  digitalWrite(CS_PIN, HIGH);
}

// Execute a command with response data e.g Read Status
void Flash_readCommand(uint8_t command, uint8_t *response, uint32_t len)
{
  digitalWrite(CS_PIN, LOW);
  _spi->write(command);
  _spi->transfer(0xFF, response, len);
  digitalWrite(CS_PIN, HIGH);
}

// Execute a command with data e.g Write Status
void Flash_writeCommand(uint8_t command, uint8_t const *data, uint32_t len)
{
  digitalWrite(CS_PIN, LOW);
  _spi->write(command);
  _spi->write(data, len);
  digitalWrite(CS_PIN, HIGH);
}

void fillAddress(uint8_t *buf, uint32_t addr)
{
  switch (_addr_len) {
  case 4:
    *buf++ = (addr >> 24) & 0xFF;
    // fall through
  case 3:
    *buf++ = (addr >> 16) & 0xFF;
    // fall through
  case 2:
  default:
    *buf++ = (addr >> 8) & 0xFF;
    *buf++ = addr & 0xFF;
  }
}
// Erase external flash by address.
// command = SFLASH_CMD_ERASE_SECTOR or SFLASH_CMD_ERASE_BLOCK
void Flash_eraseCommand(uint8_t command, uint32_t addr)
{
  digitalWrite(CS_PIN, LOW);
  uint8_t cmd_with_addr[5] = {command};
  fillAddress(cmd_with_addr + 1, addr);
  _spi->write(cmd_with_addr, 1 + _addr_len);
  digitalWrite(CS_PIN, HIGH);
}

// Read data from external flash
void Flash_readMemory(uint32_t addr, void * data, uint32_t len)
{
  digitalWrite(CS_PIN, LOW);

  uint8_t cmd_with_addr[6] = {_cmd_read};
  fillAddress(cmd_with_addr + 1, addr);

  // Fast Read has 1 extra dummy byte
  uint8_t const cmd_len = 1 + _addr_len + (SFLASH_CMD_FAST_READ == _cmd_read ? 1 : 0);
  _spi->write(cmd_with_addr, cmd_len);
  _spi->transfer((uint8_t*)data, len);

  digitalWrite(CS_PIN, HIGH);
}

// Write data to external flash. Flash sector must be previously erased.
void Flash_writeMemory(uint32_t addr, void * data, uint32_t len)
{
  digitalWrite(CS_PIN, LOW);

  uint8_t cmd_with_addr[5] = {SFLASH_CMD_PAGE_PROGRAM};
  fillAddress(cmd_with_addr + 1, addr);
  _spi->write(cmd_with_addr, 1 + _addr_len);
  _spi->transfer((const uint8_t *)data, NULL, len);

  digitalWrite(CS_PIN, HIGH);
}

//-----------------------------------------------------------------------------
typedef struct my_t {
  uint8_t u8_data;
  uint16_t u16_data;
  uint32_t u32_data;
} my_t;

typedef union page_t {
  uint8_t buffer[SFLASH_PAGE_SIZE];
  my_t my;
} page_t;

my_t rd,wr;
page_t flash_page0;
//-----------------------------------------------------------------------------
void setup()
{
  pinMode(CS_PIN, OUTPUT);
  digitalWrite(CS_PIN, HIGH);

  _spi->beginTransaction(SPISettings(40000000));

  // erease first sector 0
  Flash_eraseCommand(SFLASH_CMD_ERASE_SECTOR, 0);
  delay(100);

  // init data
  rd.u8_data = 0;
  rd.u16_data = 0;
  rd.u32_data = 0;

  wr.u8_data = 0x12;
  wr.u16_data = 0x9876;
  wr.u32_data = 0x3456ABCD;

  Serial.print("Write u8_data = "); Serial.println(wr.u8_data, HEX);
  Serial.print("Write u16_data = "); Serial.println(wr.u16_data, HEX);
  Serial.print("Write u32_data = "); Serial.println(wr.u32_data, HEX);

  // enable write
  Flash_runCommand(SFLASH_CMD_WRITE_ENABLE);
  delay(10);

  // write data to flash
  Flash_writeMemory(0, &wr, sizeof(wr));
  delay(10);

  // read back data from flash
  Flash_readMemory(0, &rd, sizeof(rd));

  Serial.print("Read u8_data = "); Serial.println(rd.u8_data, HEX);
  Serial.print("Read u16_data = "); Serial.println(rd.u16_data, HEX);
  Serial.print("Read u32_data = "); Serial.println(rd.u32_data, HEX);

}

void loop()
{
}
